from blinker import signal

started = signal("started")
stopped = signal("stopped")

message_received = signal("message_received")
message_sent = signal("message_sent")
message_consumed = signal("message_consumed")
message_error = signal("message_error")
message_drop = signal("message_drop")



from .base import BaseMiddleware
from .config import (
    BaseConfigMiddleware, PublishConfigMixin, ConsumeConfigMixin,
    NacosConfigMixin, NacosPublishConfigMiddleware, NacosConsumeConfigMiddleware,
    RedisConfigMixin, RedisPublishConfigMiddleware, RedisConsumeConfigMiddleware
)



import json
import time
import uuid
from typing import Optional, Any, Union


class Extra:
    def __init__(self, task_id=None, publish_time=None, failure_count=0):
        self.task_id = task_id or str(uuid.uuid4())
        self.publish_time = publish_time or round(time.time(), 3)
        self.failure_count = failure_count

    def to_dict(self):
        return {
            'task_id': self.task_id,
            'publish_time': self.publish_time,
            'failure_count': self.failure_count,
        }

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return f"{self.__class__.__name__}({self.task_id}, {self.publish_time}, {self.failure_count})"


class Message:

    def __init__(
            self,
            body: Optional[Union[dict, Any]] = None,
            extra: Optional[Union[dict, Extra]] = None,
            msg: Optional[Any] = None,
            broker=None
    ):
        self.body = body
        self.extra = self._set_extra(extra)
        self.msg = msg

        self.broker = broker
        self._exception = None

    @staticmethod
    def _set_extra(extra):
        if isinstance(extra, Extra):
            return extra
        elif isinstance(extra, dict):
            return Extra(**extra)
        else:
            return Extra()

    def set_exception(self, exception):
        self.exception = exception
        self.failure_count = self.failure_count + 1

    @property
    def exception(self):
        return self._exception

    @exception.setter
    def exception(self, value):
        self._exception = value

    @property
    def fail(self):
        return self.exception is not None

    @property
    def failure_count(self):
        return self.extra.failure_count

    @failure_count.setter
    def failure_count(self, value):
        self.extra.failure_count = value

    def replace(self, **kwargs):
        """替换当前message的属性"""
        for key, value in kwargs.items():
            if not hasattr(self, key):
                continue
            if key == 'extra':
                value = self._set_extra(value)
            setattr(self, key, value)
        return self

    def to_dict(self) -> dict:
        return {'body': self.body, 'extra': self.extra.to_dict()}

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def confirm(self):
        """确认消息"""
        if self.broker:
            self.broker.confirm(self)

    def reject(self):
        """拒绝消息 （原始状态重入）"""
        if self.broker:
            self.broker.reject(self)

    def requeue(self):
        """重发消息 （最新状态重入）"""
        if self.broker:
            self.broker.requeue(self)

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.body}>"



"""
将指定的函数放入线程中运行
"""
import logging
import threading
from asyncio import iscoroutinefunction
from inspect import isasyncgenfunction

from asgiref.sync import async_to_sync

from .broker import BaseBroker
from .exception import DropMessage
from .signal import message_received, message_consumed, message_error, message_drop

logger = logging.getLogger(__name__)


class WorkerThread(threading.Thread):

    def __init__(self, onestep, broker: BaseBroker, *args, **kwargs):
        """
        线程执行包装过的`onestep`函数
        :param onestep: OneStep实例
        :param broker: 监听的from broker
        """
        super().__init__(daemon=True)
        self.instance = onestep
        self.retry = self.instance.retry
        self.error_callback = self.instance.error_callback
        self.broker = broker
        self.args = args
        self.kwargs = kwargs
        self.__shutdown = False
        self.__shutdown_event = threading.Event()

    def run(self):
        """线程执行包装过的`onestep`函数

        `fn`为`onestep`函数，执行会调用`onestep`的`__call__`方法
        :return:
        """
        self.__shutdown_event.clear()

        while not self.__shutdown:
            if self.__shutdown:
                break
            # TODO：consume应当传入一些配置参数
            for message in self.broker.consume():
                if message is None:
                    continue
                message.broker = message.broker or self.broker
                logger.debug(f"receive message<{message}>")
                message_received.send(self, message=message)
                self.instance.before_emit("receive", message=message)
                self._run_instance(message)
                self.instance.after_emit("receive", message=message)

    def shutdown(self):
        self.__shutdown = True
        self.__shutdown_event.wait()

    def _run_instance(self, message):
        while True:
            try:
                if iscoroutinefunction(self.instance.fn) or isasyncgenfunction(self.instance.fn):
                    async_to_sync(self.instance)(message, *self.args, **self.kwargs)
                else:
                    self.instance(message, *self.args, **self.kwargs)
                message_consumed.send(self, message=message)
                return message.confirm()
            except DropMessage as e:
                message_drop.send(self, message=message, reason=e)
                logger.warning(f"{self.instance.fn.__name__} dropped <{type(e).__name__}: {str(e)}>")
                return message.reject()
            except Exception as e:
                message_error.send(self, message=message, error=e)
                if self.instance.state.debug:
                    logger.exception(f"{self.instance.fn.__name__} run error <{type(e).__name__}: {str(e)}>")
                else:
                    logger.error(f"{self.instance.fn.__name__} run error <{type(e).__name__}: {str(e)}>")
                message.set_exception(e)

                retry_state = self.retry(message)
                if retry_state:  # True=继续（执行重试）
                    continue
                elif retry_state is False:  # False=结束（执行回调）
                    if self.error_callback:
                        self.error_callback(message)
                    return message.reject()
                else:  # None=结束（忽略回调）
                    return message.requeue()



class BaseMiddleware:

    def before_send(self, step, message, *args, **kwargs):
        """消息发送之前"""
        pass

    def after_send(self, step, message, *args, **kwargs):
        """消息发送之后"""
        pass

    def before_receive(self, step, message, *args, **kwargs):
        """消息接收之前"""
        pass

    def after_receive(self, step, message, *args, **kwargs):
        """消息接收之后"""
        pass



from .base import (
    BaseBroker, BaseConsumer, BaseLocalBroker, BaseLocalConsumer
)
from .memory import MemoryBroker
from .webhook import WebHookBroker
from .rabbitmq import RabbitMQBroker
from .cron import CronBroker



import functools
import collections
import logging

from inspect import isgenerator, iscoroutinefunction, isasyncgenfunction, isasyncgen
from typing import Optional, List, Dict, Any, Callable, Union

from .broker.base import BaseBroker
from .exception import StopMiddleware
from .message import Message
from .retry import NeverRetry
from .signal import message_sent, started, stopped
from .state import State
from .worker import WorkerThread

logger = logging.getLogger(__name__)

DEFAULT_WORKERS = 1


class BaseOneStep:
    consumers: Dict[str, List[WorkerThread]] = collections.defaultdict(list)
    state = State()  # 全局状态

    def __init__(self, fn,
                 group: str = "OneStep",
                 from_broker: Union[BaseBroker, List[BaseBroker], None] = None,
                 to_broker: Union[BaseBroker, List[BaseBroker], None] = None,
                 workers: Optional[int] = None,
                 middlewares: Optional[List[Any]] = None,
                 retry: Union[Callable, object] = NeverRetry(),
                 error_callback: Optional[Union[Callable, object]] = None):
        self.group = group
        self.fn = fn
        self.workers = workers or DEFAULT_WORKERS
        self.middlewares = middlewares or []

        self.from_brokers = self._init_broker(from_broker)
        self.to_brokers = self._init_broker(to_broker)
        self.retry = retry
        self.error_callback = error_callback

        for broker in self.from_brokers:
            self._add_consumer(broker)

    @staticmethod
    def _init_broker(broker: Union[BaseBroker, List[BaseBroker], None] = None):
        if not broker:
            return []

        if isinstance(broker, BaseBroker):
            return [broker]
        if isinstance(broker, (list, tuple)):
            return list(broker)
        raise TypeError(
            f"broker must be BaseBroker or list or tuple, not {type(broker)}")

    def _add_consumer(self, broker):
        for _ in range(self.workers):
            self.consumers[self.group].append(
                WorkerThread(onestep=self, broker=broker)
            )

    @classmethod
    def _find_consumers(cls, group: Optional[str] = None):
        """按组查找消费者"""
        if group is None:
            consumers = [c for v in cls.consumers.values() for c in v]
        else:
            consumers = cls.consumers[group]
        return consumers

    @classmethod
    def start(cls, group: Optional[str] = None):
        for consumer in cls._find_consumers(group):
            consumer.start()

    @classmethod
    def stop(cls, group: Optional[str] = None):
        for consumer in cls._find_consumers(group):
            consumer.shutdown()

    def wraps(self, func):
        @functools.wraps(func)
        def wrapped_f(*args, **kwargs):
            return self(*args, **kwargs)  # noqa

        return wrapped_f

    def send(self, result, broker=None):
        """将返回的内容交给broker发送"""
        if result is None:
            logger.debug("send(result): body is empty")
            return

        brokers = self._init_broker(broker) or self.to_brokers
        if not brokers:
            logger.debug("send(result): broker is empty")
            return

        # 如果是Message类型，就不再封装
        message = result if isinstance(result, Message) else Message(body=result)

        self.before_emit("send", message=message)
        for broker in brokers:
            message_sent.send(self, message=message, broker=broker)
            broker.send(message)
        self.after_emit("send", message=message)

    def before_emit(self, signal, *args, **kwargs):
        signal = "before_" + signal
        self.emit(signal, *args, **kwargs)

    def after_emit(self, signal, *args, **kwargs):
        signal = "after_" + signal
        self.emit(signal, *args, **kwargs)

    def emit(self, signal, *args, **kwargs):
        for middleware in self.middlewares:
            if not hasattr(middleware, signal):
                continue
            try:
                getattr(middleware, signal)(step=self, *args, **kwargs)
            except StopMiddleware:
                break


def decorator_func_proxy(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


class SyncOneStep(BaseOneStep):

    def __call__(self, *args, **kwargs):
        """同步执行原函数"""

        result = self.fn(*args, **kwargs)

        if isgenerator(result):
            for item in result:
                self.send(item)
        else:
            self.send(result)
        return result


class AsyncOneStep(BaseOneStep):

    async def __call__(self, *args, **kwargs):
        """"异步执行原函数"""

        if iscoroutinefunction(self.fn):
            result = await self.fn(*args, **kwargs)
        else:
            result = self.fn(*args, **kwargs)

        if isasyncgen(result):
            async for item in result:
                self.send(item)
        else:
            self.send(result)
        return result


class step:

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func, *_args, **_kwargs):
        if iscoroutinefunction(func) or isasyncgenfunction(func):
            os = AsyncOneStep(fn=func, *self.args, **self.kwargs)
        else:
            os = SyncOneStep(fn=func, *self.args, **self.kwargs)

        return os.wraps(func)

    @staticmethod
    def start(group=None, block=None):
        BaseOneStep.start(group=group)
        started.send()
        if block:
            import time
            while True:
                time.sleep(1)

    @staticmethod
    def stop(group=None):
        BaseOneStep.stop(group=group)
        stopped.send()

    @staticmethod
    def set_debugging():

        if not BaseOneStep.state.debug:
            onestep_logger = logging.getLogger("onestep")
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter(
                "[%(levelname)s] %(asctime)s %(name)s:%(message)s"))
            onestep_logger.addHandler(handler)
            onestep_logger.setLevel(logging.DEBUG)
            BaseOneStep.state.debug = True



class StopMiddleware(Exception):
    ...


class RetryException(Exception):
    def __init__(self, message=None, times=None, **kwargs):
        self.message = message
        self.times = times
        self.kwargs = kwargs


class RetryViaQueue(RetryException):
    """消息重试-通过重试队列

    抛出此异常，消息将被重新放入队列，等待下次消费。

    具体实现是触发`broker`中`nack`方法，`requeue`参数为True，将消息重新放入队列
    """


class RetryViaLocal(RetryException):
    """消息重试-本地

    不经过队列，直接在本地重试，直到达到重试次数。
    """


class DropMessage(Exception):
    """从 Brokers 中 丢弃该消息"""



from abc import ABC, abstractmethod
from typing import Optional, Tuple
from .exception import RetryViaLocal, RetryViaQueue
from .message import Message


class BaseRetry(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Optional[bool]:
        """True=继续（执行重试）， False=结束（执行回调），None=结束（忽略回调）"""
        pass


class NeverRetry(BaseRetry):

    def __call__(self, message) -> Optional[bool]:
        return False


class AlwaysRetry(BaseRetry):

    def __call__(self, message) -> Optional[bool]:
        return True


class TimesRetry(BaseRetry):

    def __init__(self, times: int = 3):
        self.times = times

    def __call__(self, message) -> Optional[bool]:
        return message.failure_count < self.times


class RetryIfException(BaseRetry):

    def __init__(self, exceptions: Optional[Tuple[Exception]] = Exception):
        self.exceptions = exceptions

    def __call__(self, message) -> Optional[bool]:
        return isinstance(message.exception, self.exceptions)  # noqa


class LocalAndQueueRetry(TimesRetry):

    def __call__(self, message: Message) -> Optional[bool]:
        """本地重试或队列重试"""
        if isinstance(message.exception, (RetryViaLocal, RetryViaQueue)):
            if message.failure_count < (message.exception.times or self.times):
                if isinstance(message.exception, RetryViaQueue):
                    return None  # 入队重试，不回调
                return True  # 本地重试，不回调
            else:
                return False  # 不重试了，回调
        else:
            return False  # 其他异常，回调


# error callback
class BaseErrorCallback(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class NackErrorCallBack(BaseErrorCallback):

    def __call__(self, message):
        message.reject()



# -*- coding: utf-8 -*-
import abc
import json
from queue import Queue, Empty

from ..exception import StopMiddleware
from ..message import Message


class BaseBroker:

    def __init__(self, name=None, queue=None, middlewares=None):
        self.queue = queue
        self.name = name or "broker"
        self.middlewares = []

        if middlewares:
            for middleware in middlewares:
                self.add_middleware(middleware)

    def add_middleware(self, middleware):
        self.middlewares.append(middleware)

    def send(self, message):
        """对消息进行预处理，然后再发送"""
        if not isinstance(message, Message):
            message = Message(body=message)
        # TODO: 对消息发送进行N次重试，确保消息发送成功。
        return self.publish(message.to_json())

    @abc.abstractmethod
    def publish(self, message):
        """
        如果当前Broker是Job的to_broker, 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def consume(self, *args, **kwargs):
        """
        如果当前Broker是Job的from_broker, 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def confirm(self, message):
        """确认消息"""
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def reject(self, message):
        """拒绝消息"""
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def requeue(self, message, is_source=False):
        """
        重发消息：先拒绝 再 重入
        is_source = False 重入使用消息的当前状态
        is_source = True 重入使用消息的初始状态
        """
        raise NotImplementedError('Please implement in subclasses.')

    def before_emit(self, signal, *args, **kwargs):
        signal = "before_" + signal
        self._emit(signal, *args, **kwargs)

    def after_emit(self, signal, *args, **kwargs):
        signal = "after_" + signal
        self._emit(signal, *args, **kwargs)

    def _emit(self, signal, *args, **kwargs):
        for middleware in self.middlewares:
            if not hasattr(middleware, signal):
                continue
            try:
                getattr(middleware, signal)(self, *args, **kwargs)
            except StopMiddleware:
                break

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self):
        return self.name


class BaseConsumer:

    def __init__(self, queue: Queue, *args, **kwargs):
        self.queue = queue
        self.timeout = kwargs.pop("timeout", 1000)

    @abc.abstractmethod
    def _to_message(self, data):
        """
        转换消息内容到 Message , 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    def __next__(self):
        try:
            return self._to_message(self.queue.get(timeout=self.timeout / 1000))
        except Empty:
            return None

    def __iter__(self):
        return self


class BaseLocalBroker(BaseBroker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = Queue()

    def publish(self, message):
        self.queue.put_nowait(message)

    def consume(self, *args, **kwargs):
        return BaseLocalConsumer(self.queue, *args, **kwargs)

    def confirm(self, message):
        """确认消息"""
        pass

    def reject(self, message):
        """拒绝消息"""
        pass

    def requeue(self, message, is_source=False):
        """重发消息：先拒绝 再 重入"""
        if is_source:
            self.publish(message.msg)
        else:
            self.send(message)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self):
        return self.name


class BaseLocalConsumer(BaseConsumer):

    def _to_message(self, data):
        message = Message(msg=data)
        try:
            body = json.loads(data)
            message.replace(**body)
        except (json.JSONDecodeError, TypeError):
            message.body = data
        return message



# -*- coding: utf-8 -*-
import typing


class State:
    _state: typing.Dict[str, typing.Any]

    def __init__(self, state: typing.Optional[typing.Dict[str, typing.Any]] = None):
        if state is None:
            state = {}
        super().__setattr__("_state", state)

    def __setattr__(self, key: typing.Any, value: typing.Any) -> None:
        self._state[key] = value

    def __getattr__(self, key: typing.Any) -> typing.Any:
        try:
            return self._state[key]
        except KeyError:
            return None

    def __delattr__(self, key: typing.Any) -> None:
        del self._state[key]



import logging
from threading import local
import amqpstorm
from amqpstorm.exception import AMQPConnectionError

MAX_SEND_ATTEMPTS = 6  # 最大发送重试次数
MAX_CONNECTION_ATTEMPTS = float('inf')  # 最大连接重试次数

logger = logging.Logger(__name__)


class RabbitmqStore:

    def __init__(self, *, confirm_delivery=True, host=None, port=None, username=None, password=None,
                 **kwargs):
        """
        :param confirm_delivery: 是否开启消息确认
        :param host: RabbitMQ host
        :param port: RabbitMQ port
        :param username: RabbitMQ username
        :param password: RabbitMQ password
        :param kwargs: RabbitMQ parameters
        """
        self.state = local()
        self.parameters = {
            'hostname': host or 'localhost',
            'port': port or 5672,
            'username': username or 'guest',
            'password': password or 'guest',
        }
        if kwargs:
            self.parameters.update(kwargs)
        self.confirm_delivery = confirm_delivery

    def _create_connection(self):
        attempts = 1
        while attempts <= MAX_CONNECTION_ATTEMPTS:
            try:
                return amqpstorm.Connection(**self.parameters)
            except AMQPConnectionError as exc:
                attempts += 1
                logger.warning("RabbitmqStore connection error: %s", exc)
        raise AMQPConnectionError("RabbitmqStore connection error, max attempts reached")

    @property
    def connection(self):
        connection = getattr(self.state, "connection", None)
        if connection is None or not connection.is_open:
            connection = self.state.connection = self._create_connection()
        return connection

    @connection.deleter
    def connection(self):
        if _connection := getattr(self.state, "connection", None):
            _connection.close()
            del self.state.connection
            del self.state.channel

    @property
    def channel(self):
        connection = getattr(self.state, "connection", None)
        channel = getattr(self.state, "channel", None)
        if all([connection, channel]) and all([connection.is_open, channel.is_open]):
            return channel
        channel = self.state.channel = self.connection.channel()
        if self.confirm_delivery:
            channel.confirm_deliveries()
        return channel

    def declare_queue(self, queue_name, arguments=None):
        """声明队列"""
        if arguments is None:
            arguments = {}
        return self.channel.queue.declare(queue_name, durable=True, arguments=arguments)

    def send(self, queue_name, message, priority=None, **kwargs):
        """发送消息"""
        attempts = 1
        while True:
            try:
                self.declare_queue(queue_name)
                self.channel.basic.publish(
                    message, queue_name, properties=priority, **kwargs
                )
                return message
            except Exception as exc:
                del self.connection
                attempts += 1
                if attempts > MAX_SEND_ATTEMPTS:
                    raise exc

    def flush_queue(self, queue_name):
        """清空队列"""
        self.channel.queue.purge(queue_name)

    def get_message_counts(self, queue_name: str) -> int:
        """获取消息数量"""
        queue_response = self.declare_queue(queue_name)
        return queue_response.get("message_count", 0)

    def start_consuming(self, queue_name, callback, prefetch=1, **kwargs):
        """开始消费"""
        while True:
            try:
                self.channel.basic.qos(prefetch_count=prefetch)
                self.channel.basic.consume(queue=queue_name, callback=callback, no_ack=False, **kwargs)
                self.channel.start_consuming(to_tuple=False)
            except AMQPConnectionError:
                logger.warning("RabbitmqStore consume connection error, reconnecting...")
                del self.connection
            except Exception as e:
                logger.exception(f"RabbitmqStore consume error<{e}>, reconnecting...")
                del self.connection



import logging
import threading
import collections
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer

from .base import BaseLocalBroker, BaseLocalConsumer

logger = logging.getLogger(__name__)

Server = collections.namedtuple("Server", ["path", "queue"])


class WebHookServer(BaseHTTPRequestHandler):
    servers = collections.defaultdict(list)

    def do_POST(self):
        """
        接收WebHook请求
        """
        server_paths = WebHookServer.servers.get(self.server.server_address, [])
        for server in server_paths:
            if self.path == server.path:
                queue = server.queue
                break
        else:
            return self.send_error(404)

        content_len = int(self.headers.get('content-length', 0))
        post_body = self.rfile.read(content_len).decode("utf-8")
        queue.put_nowait(post_body)
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        self.wfile.write(b'{ "status": "ok" }')


class WebHookBroker(BaseLocalBroker):
    _servers = {}

    def __init__(self,
                 path: str,
                 host: str = "0.0.0.0",
                 port: int = 8090,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.host = host
        self.port = port
        self.path = path

        self._create_server()
        logger.debug(f"WebHookBroker: {self.host}:{self.port}{self.path}")

    def _create_server(self):

        if (self.host, self.port) not in self._servers:
            hs = ThreadingHTTPServer(
                (self.host, self.port),
                WebHookServer
            )
            self._servers[(self.host, self.port)] = hs
        else:
            hs = self._servers[(self.host, self.port)]

        WebHookServer.servers[(self.host, self.port)].append(Server(self.path, self.queue))
        threading.Thread(target=hs.serve_forever).start()

    def consume(self, *args, **kwargs):
        return WebHookConsumer(self.queue, *args, **kwargs)


class WebHookConsumer(BaseLocalConsumer):
    ...



from .onestep import step
from .retry import NeverRetry, AlwaysRetry, TimesRetry, RetryIfException, NackErrorCallBack, LocalAndQueueRetry
from .broker import (
    BaseBroker, BaseConsumer, BaseLocalBroker, BaseLocalConsumer,
    MemoryBroker, RabbitMQBroker, WebHookBroker, CronBroker
)
from .middleware import (
    BaseMiddleware, BaseConfigMiddleware,
    NacosPublishConfigMiddleware, NacosConsumeConfigMiddleware,
    RedisPublishConfigMiddleware, RedisConsumeConfigMiddleware,
)

__all__ = [
    'step',

    # broker
    'BaseBroker',
    'BaseConsumer',
    'BaseLocalBroker',
    'BaseLocalConsumer',
    'MemoryBroker',
    'RabbitMQBroker',
    'WebHookBroker',
    'CronBroker',

    # retry
    'NeverRetry',
    'AlwaysRetry',
    'TimesRetry',
    'RetryIfException',
    'LocalAndQueueRetry',
    #
    'NackErrorCallBack',

    # middleware
    'BaseMiddleware',
    'BaseConfigMiddleware',
    'NacosPublishConfigMiddleware',
    'NacosConsumeConfigMiddleware',
    'RedisPublishConfigMiddleware',
    'RedisConsumeConfigMiddleware',
]



import json
import threading
from queue import Queue
from typing import Optional, Dict

import amqpstorm

from .base import BaseBroker, BaseConsumer
from ..store.rabbitmq import RabbitmqStore
from ..message import Message


class RabbitMQBroker(BaseBroker):

    def __init__(self, queue_name, params: Optional[Dict] = None, prefetch: Optional[int] = 1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue_name = queue_name
        self.queue = Queue()
        params = params or {}
        self.client = RabbitmqStore(**params)
        self.client.declare_queue(self.queue_name)
        self.prefetch = prefetch

    def _consume(self, *args, **kwargs):
        def callback(message):
            self.queue.put(message)

        prefetch = kwargs.pop("prefetch", self.prefetch)
        self.client.start_consuming(queue_name=self.queue_name, callback=callback, prefetch=prefetch, **kwargs)

    def consume(self, *args, **kwargs):
        threading.Thread(target=self._consume, *args, **kwargs).start()
        return RabbitMQConsumer(self.queue)

    def publish(self, message):
        self.client.send(self.queue_name, message)

    def confirm(self, message):
        """确认消息"""
        message.msg.ack()

    def reject(self, message):
        """拒绝消息"""
        message.msg.nack(requeue=False)

    def requeue(self, message, is_source=False):
        """重发消息：先拒绝 再 重入"""
        if is_source:
            message.msg.nack(requeue=True)
        else:
            message.msg.nack(requeue=False)
            self.send(message)


class RabbitMQConsumer(BaseConsumer):
    def _to_message(self, data: amqpstorm.Message):
        try:
            message = json.loads(data.body)
        except json.JSONDecodeError:
            message = {"body": data.body}
        if not isinstance(message, dict):
            message = {"body": message}

        return Message(body=message.get("body"), extra=message.get("extra"), msg=data)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['onestep', 'onestep.broker', 'onestep.middleware', 'onestep.store']

package_data = \
{'': ['*']}

install_requires = \
['asgiref>=3.6.0,<4.0.0', 'blinker>=1.5,<2.0', 'croniter>=1.3.8,<2.0.0']

extras_require = \
{'rabbitmq': ['amqpstorm>=2.10.6,<3.0.0']}

setup_kwargs = {
    'name': 'onestep',
    'version': '0.1.5',
    'description': '',
    'long_description': '# OneStep\n\n<a href="https://github.com/mic1on/onestep/actions/workflows/test.yml?query=event%3Apush+branch%3Amain" target="_blank">\n    <img src="https://github.com/mic1on/onestep/workflows/test%20suite/badge.svg?branch=main&event=push" alt="Test">\n</a>\n<a href="https://pypi.org/project/onestep" target="_blank">\n    <img src="https://img.shields.io/pypi/v/onestep.svg" alt="Package version">\n</a>\n\n<a href="https://pypi.org/project/onestep" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/onestep.svg" alt="Supported Python versions">\n</a>\n\n<hr />\n仅需一步，轻松实现分布式异步任务。\n\n## Brokers\n\n- [x] MemoryBroker\n- [x] CronBroker\n- [x] WebHookBroker\n- [x] RabbitMQBroker\n- [ ] RedisBroker\n- [ ] KafkaBroker\n\n## example\n\n```python\nfrom onestep import step, WebHookBroker\n\n\n# 对外提供一个webhook接口，接收外部的消息\n@step(from_broker=WebHookBroker(path="/push"))\ndef waiting_messages(message):\n    print("收到消息：", message)\n\n\nif __name__ == \'__main__\':\n    step.start(block=True)\n```\n\n```python\nfrom onestep import step, CronBroker\n\n\n# 每3秒触发一次任务\n@step(from_broker=CronBroker("* * * * * */3", a=1))\ndef cron_task(message):\n    assert message.body == {"a": 1}\n    return message\n\n\nif __name__ == \'__main__\':\n    step.start(block=True)\n```\n\n更多例子请参阅：[examples](example)',
    'author': 'miclon',
    'author_email': 'jcnd@163.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .base import BaseMiddleware


class BaseConfigMiddleware(BaseMiddleware):

    def __init__(self, client, *args, **kwargs):
        self.config_key = kwargs.pop('config_key', None)
        self.client = client

    def process(self, step, message, *args, **kwargs):
        """实际获取配置的逻辑"""
        config = self.get(self.config_key)
        return config

    def get(self, key):
        raise NotImplementedError


class PublishConfigMixin:
    def before_receive(self, step, message, *args, **kwargs):
        """消息发送之前，给消息添加配置"""
        config = self.process(step, message, *args, **kwargs)  # noqa
        # 持久性，会跟随消息传递到其他 broker
        message.extra['config'] = config


class ConsumeConfigMixin:
    def before_receive(self, step, message, *args, **kwargs):
        """消息消费之前，给消息附加配置"""
        config = self.process(step, message, *args, **kwargs)  # noqa
        # 一次性，不跟随消息传递到其他 broker
        message.config = config


class NacosConfigMixin:

    def __init__(self, client, *args, **kwargs):
        super().__init__(client, *args, **kwargs)
        self.config_group = kwargs.pop("config_group", "DEFAULT_GROUP")

    def get(self, key):
        pass

    def set(self, key, value):
        pass


class NacosPublishConfigMiddleware(NacosConfigMixin, PublishConfigMixin, BaseConfigMiddleware):
    """发布前附加来自 Nacos 的配置"""


class NacosConsumeConfigMiddleware(NacosConfigMixin, ConsumeConfigMixin, BaseConfigMiddleware):
    """消费前附加来自 Nacos 的配置"""


class RedisConfigMixin:

    def __init__(self, client, *args, **kwargs):
        super().__init__(client, *args, **kwargs)
        self.config_group = kwargs.pop('config_group', 'onestep:config')

    def get(self, key):
        value = self.client.hget(name=self.config_group, key=key)  # noqa
        if isinstance(value, bytes):
            value = value.decode()
        return value

    def set(self, key, value):
        return self.client.hset(name=self.config_group, key=key, value=value)  # noqa


class RedisPublishConfigMiddleware(RedisConfigMixin, PublishConfigMixin, BaseConfigMiddleware):
    """发布前附加来自 Redis 的配置"""


class RedisConsumeConfigMiddleware(RedisConfigMixin, ConsumeConfigMixin, BaseConfigMiddleware):
    """消费前附加来自 Redis 的配置"""



"""
使用CRON表达式触发任务执行
"""
import threading
from datetime import datetime

from croniter import croniter

from .base import BaseLocalBroker, BaseLocalConsumer


class CronBroker(BaseLocalBroker):

    def __init__(self, cron, name=None, middlewares=None, **kwargs):
        super().__init__(name=name, middlewares=middlewares)
        self.cron = cron
        self.itr = croniter(cron, datetime.now())
        self.next_fire_time = self.itr.get_next(datetime)
        self.kwargs = kwargs
        self._scheduler()

    def _scheduler(self):
        if self.next_fire_time <= datetime.now():
            self.next_fire_time = self.itr.get_next(datetime)
            self.queue.put_nowait(self.kwargs)

        threading.Timer(interval=1, function=self._scheduler).start()

    def consume(self):
        return CronConsumer(self.queue)


class CronConsumer(BaseLocalConsumer):
    ...
